# def test_schema_current_version():
#    import lnschema_core  # noqa
#
#    current_version = lnschema_core.__version__
#
#    print(100 * "*", current_version)
#    if current_version == "0.3.1":
#        get_hub_with_authentication()
#
#
# def test_schema_old_version():
#    import lnschema_core  # noqa
#
#    current_version = lnschema_core.__version__
#
#    print(100 * "*", current_version)
#    if current_version == "0.3.0":
#        with pytest.raises(RuntimeError):
#            get_hub_with_authentication()
#
